// -*- C++ -*-
/*!
 * @file  FireWebCamera.cpp
 * @brief Capture Image from PointGray IEEE1394 Camera and Web Camera
 * @date $Date$
 *
 * $Id$
 */

#include "FireWebCamera.h"
#include "CfgFileLoader.h"

// Module specification
static const char* firewebcamera_spec[] =
  {
    "implementation_id", "FireWebCamera",
    "type_name",         "FireWebCamera",
    "description",       "Capture Image from PointGray IEEE1394 Camera and Web Camera",
    "version",           "0.5.0",
    "vendor",            "Hideyasu Takahashi , Arai Lab.,Osaka Univ.",
    "category",          "Camera Controller",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    "exec_cxt.periodic.rate", "1000.0",
    // Configuration variables
    "conf.default.CameraModels", "WebCamera",
    "conf.default.PixelFormat", "RGB",
	"conf.default.ImageWindow", "off",
    ""
  };


/*!
 * @brief constructor
 * @param manager Maneger Object
 */
FireWebCamera::FireWebCamera(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_ImageDataOut("ImageData", m_ImageData),
	m_ImageDataROut("ImageDataR", m_ImageDataR),
    m_CamInfoPort("CamInfo"),
	m_CamInfoRPort("CamInfoR"),
	m_StereoInfoPort("StereoInfo")
{
  // Registration: InPort/OutPort/Service
  // Set InPort buffers
  
  // Set OutPort buffer
  registerOutPort("ImageData", m_ImageDataOut);
  registerOutPort("ImageDataR", m_ImageDataROut);
  
  // Set service provider to Ports
  m_CamInfoPort.registerProvider("ImageInfoService", "ImageInfoService", m_ImageInfoService);
  m_CamInfoPort.registerProvider("CameraInfoService", "CameraInfoService", m_CameraInfoService);
  m_CamInfoRPort.registerProvider("ImageInfoServiceR", "ImageInfoService", m_ImageInfoServiceR);
  m_CamInfoRPort.registerProvider("CameraInfoServiceR", "CameraInfoService", m_CameraInfoServiceR);
  m_StereoInfoPort.registerProvider("StereoInfoService", "StereoCameraInfoService", m_StereoInfoService);
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  registerPort(m_CamInfoPort);
  registerPort(m_CamInfoRPort);
  registerPort(m_StereoInfoPort);
}

/*!
 * @brief destructor
 */
FireWebCamera::~FireWebCamera()
{
}



RTC::ReturnCode_t FireWebCamera::onInitialize()
{
	// Bind variables and configuration variable
	bindParameter("CameraModels", m_CameraModels, "WebCamera");
	bindParameter("PixelFormat", m_PixelFormat, "RGB");
	bindParameter("ImageWindow", m_ImageWindow, "off");

	// set window's name
	_imgViewer[0].setWindowName("FireWebCamera Image");
	_imgViewer[1].setWindowName("FireWebCamera Right Image");

	// lock all service port
	m_ImageInfoService.lock_getImageHeightLock();
	m_ImageInfoService.lock_getImageWidthLock();
	m_ImageInfoService.lock_getPixelFormatLock();
	m_ImageInfoServiceR.lock_getImageHeightLock();
	m_ImageInfoServiceR.lock_getImageWidthLock();
	m_ImageInfoServiceR.lock_getPixelFormatLock();
	m_CameraInfoService.lock_getDistortionService();
	m_CameraInfoService.lock_getIntrinsicService();
	m_CameraInfoServiceR.lock_getDistortionService();
	m_CameraInfoServiceR.lock_getIntrinsicService();
	m_StereoInfoService.lock_getExtrinsicService();

	std::cout << "### FireWebCamera Component ###" << std::endl;

  return RTC::RTC_OK;
}


RTC::ReturnCode_t FireWebCamera::onFinalize()
{
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t FireWebCamera::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t FireWebCamera::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t FireWebCamera::onActivated(RTC::UniqueId ec_id)
{
	// select camera model
	bool webcam(true), stereo(false);
	if(m_CameraModels=="WebCamera"){
		std::cout << "camera: Web Camera" << std::endl;
	}else if(m_CameraModels=="IEEE1394Camera"){
		webcam = false;
		std::cout << "camera: IEEE1394 Camera" << std::endl;
	}else if(m_CameraModels=="IEEE1394StereoCamera"){
		webcam = false;
		stereo = true;
		std::cout << "camera: IEEE1394 Stereo Camera" << std::endl;
	}

	// select pixel format
	if(m_PixelFormat=="RGB"){
		_camera.setPixelType(pixel_t::PIXEL_RGB);
	}else if(m_PixelFormat=="BGR"){
		_camera.setPixelType(pixel_t::PIXEL_BGR);
	}

	// open camera
	if(!_camera.Opens(webcam, stereo)){
		std::cout << "Open camera: failed" << std::endl;
		return RTC::RTC_ERROR;
	}
	std::cout << "Open camera: done" << std::endl;

	// get image size and pixel format from camera, and set them
	// set image size
	int iwidth, iheight;
	_camera.getImageSize(&iwidth, &iheight);
	m_ImageInfoService.setImageHeight(iheight);
	m_ImageInfoServiceR.setImageHeight(iheight);
	m_ImageInfoService.setImageWidth(iwidth);
	m_ImageInfoServiceR.setImageWidth(iwidth);

	// unlock service port function
	//    : getImageHeight()
	//    : getImageWidth()
	m_ImageInfoService.unlock_getImageHeightLock();
	m_ImageInfoServiceR.unlock_getImageHeightLock();
	m_ImageInfoService.unlock_getImageWidthLock();
	m_ImageInfoServiceR.unlock_getImageWidthLock();
	
	std::cout << "Image Size: height " << iheight << ",width " << iwidth << std::endl;

	// set pixel format
	m_ImageInfoService.setImagePixelFormat((short)_camera.getPixelType());
	m_ImageInfoServiceR.setImagePixelFormat((short)_camera.getPixelType());

	// unlock service port function
	//    : getPixelFormat()
	m_ImageInfoService.unlock_getPixelFormatLock();
	m_ImageInfoServiceR.unlock_getPixelFormatLock();

	// create image buffer
	int isize = iwidth*iheight*3; // width*height*3channels
	m_ImageData.data.length(isize);
	if(stereo){
		m_ImageDataR.data.length(isize);
	}

	// read intrinsic parameter and distortion parameter from configuration file
	Cfg::CfgFileLoader<double> cfg;
	cfg.setNewCommand("CameraParameter.Intrinsic1", 3);
	cfg.setNewCommand("CameraParameter.Intrinsic2", 3);
	cfg.setNewCommand("CameraParameter.Distortion", 5);
	if(stereo){
		cfg.setNewCommand("RightCameraParameter.Intrinsic1", 3);
		cfg.setNewCommand("RightCameraParameter.Intrinsic2", 3);
		cfg.setNewCommand("RightCameraParameter.Distortion", 5);
		cfg.setNewCommand("StereoCameraParameter.Baseline", 1);
	}

	// load configuration file
	if(cfg.loadCfgFile("camera.cfg")){
		std::cout << "load camera parameter from cfg file: done" << std::endl;
		// load left camera parameter
		std::vector<double> p;
		double intrinsic[6], distortion[5], baseline;
		p = cfg.getParameter("CameraParameter.Intrinsic1");
		for(int i=0; i<p.size(); ++i){intrinsic[i] = p[i];}
		std::cout << "   -CameraParameter.Intrinsic1:" << p[0] << ", " << p[1] << ", " << p[2] << std::endl;
		p = cfg.getParameter("CameraParameter.Intrinsic2");
		for(int i=0; i<p.size(); ++i){intrinsic[i+3] = p[i];}
		std::cout << "   -CameraParameter.Intrinsic2:" << p[0] << ", " << p[1] << ", " << p[2] << std::endl;
		p = cfg.getParameter("CameraParameter.Distortion");
		for(int i=0; i<p.size(); ++i){distortion[i] = p[i];}
		std::cout << "   -CameraParameter.Distortion:" << std::ends;
		for(int i=0; i<p.size(); ++i){std::cout << p[i] << ", " << std::ends;} std::cout << "" << std::endl;
		// set camera parameter
		m_CameraInfoService.setIntrinsicParameter(intrinsic);
		m_CameraInfoService.setDistortionParameter(distortion);

		if(stereo){
			// load right camera parameter
			p = cfg.getParameter("RightCameraParameter.Intrinsic1");
			for(int i=0; i<p.size(); ++i){intrinsic[i] = p[i];}
			std::cout << "   -RightCameraParameter.Intrinsic1:" << intrinsic[0] << ", " << intrinsic[1] << ", " << intrinsic[2] << std::endl;
			p = cfg.getParameter("RightCameraParameter.Intrinsic2");
			for(int i=0; i<p.size(); ++i){intrinsic[i+3] = p[i];}
			std::cout << "   -RightCameraParameter.Intrinsic2:" << intrinsic[3] << ", " << intrinsic[4] << ", " << intrinsic[5] << std::endl;
			p = cfg.getParameter("RightCameraParameter.Distortion");
			for(int i=0; i<p.size(); ++i){distortion[i] = p[i];}
			std::cout << "   -RightCameraParameter.Distortion:" <<std::ends;
			for(int i=0; i<p.size(); ++i){std::cout << distortion[i] << ", " << std::ends;} std::cout << "" << std::endl;
			p = cfg.getParameter("StereoCameraParameter.Baseline");
			for(int i=0; i<p.size(); ++i){baseline = p[i];}
			std::cout << "   -StereoCameraParameter.Baseline:" << baseline << std::endl;
			// set right camera parameter
			m_CameraInfoServiceR.setIntrinsicParameter(intrinsic);
			m_CameraInfoServiceR.setDistortionParameter(distortion);
			m_StereoInfoService.setBaseLine(baseline);

			// unlock service port function
			m_CameraInfoServiceR.unlock_getDistortionService();
			m_CameraInfoServiceR.unlock_getIntrinsicService();
			m_StereoInfoService.unlock_getExtrinsicService();
		}

		// unlock service port function
		m_CameraInfoService.unlock_getDistortionService();
		m_CameraInfoService.unlock_getIntrinsicService();
	}else{ // configulation file does not exist
		std::cout << "load camera parameter :failed" << std::endl;
		std::cout << "camera.cfg does not exist." << std::endl;
	}

	// set window size
	for(int i=0; i<2; ++i){
		_imgViewer[i].setWindowSize(iwidth, iheight);
	}
  return RTC::RTC_OK;
}


RTC::ReturnCode_t FireWebCamera::onDeactivated(RTC::UniqueId ec_id)
{
	// close camera
	if(!_camera.Closes()){
		std::cout << "Close camera :failed" << std::endl;
		return RTC::RTC_ERROR;
	}
	std::cout << "Close camera :done " << std::endl;

    // close window
	for (int i=0; i<2; ++i){
		_imgViewer[i].closeWindow();
	}

	// lock all service port
	m_ImageInfoService.lock_getImageHeightLock();
	m_ImageInfoService.lock_getImageWidthLock();
	m_ImageInfoService.lock_getPixelFormatLock();
	m_ImageInfoServiceR.lock_getImageHeightLock();
	m_ImageInfoServiceR.lock_getImageWidthLock();
	m_ImageInfoServiceR.lock_getPixelFormatLock();
	m_CameraInfoService.lock_getDistortionService();
	m_CameraInfoService.lock_getIntrinsicService();
	m_CameraInfoServiceR.lock_getDistortionService();
	m_CameraInfoServiceR.lock_getIntrinsicService();
	m_StereoInfoService.lock_getExtrinsicService();
	
  return RTC::RTC_OK;
}


RTC::ReturnCode_t FireWebCamera::onExecute(RTC::UniqueId ec_id)
{
	// select pixel format
	// lock service
	m_ImageInfoService.lock_getPixelFormatLock();
	m_ImageInfoServiceR.lock_getPixelFormatLock();
	if(m_PixelFormat=="RGB"){_camera.setPixelType(pixel_t::PIXEL_RGB);}
	else if(m_PixelFormat=="BGR"){_camera.setPixelType(pixel_t::PIXEL_BGR);}
	m_ImageInfoService.setImagePixelFormat((short)_camera.getPixelType());
	m_ImageInfoServiceR.setImagePixelFormat((short)_camera.getPixelType());
	// unlock service
	m_ImageInfoService.unlock_getPixelFormatLock();
	m_ImageInfoServiceR.unlock_getPixelFormatLock();

	if(_camera.getStereo()) {// for stereo camera
		// capture image
		if(!_camera.CaptureImages((unsigned char*)&(m_ImageData.data[0]), (unsigned char*)&(m_ImageDataR.data[0]))) {
			return RTC::RTC_ERROR;
		}

		// time stamp
		m_ImageData.tm = UTIL::getCurrentTimeStamp();
		m_ImageDataR.tm = m_ImageData.tm;

		// send image to outport
		if(!m_ImageDataOut.write()) {
			std::cout << "Left image was not sent!" << std::endl;
			std::cout << "TimeStamp: " << m_ImageData.tm.sec << "[sec] " << m_ImageData.tm.nsec << "[nsec]" << std::endl;
		}
		if(!m_ImageDataROut.write()) {
			std::cout << "Right image was not sent!" << std::endl;
			std::cout << "TimeStamp: " << m_ImageData.tm.sec << "[sec] " << m_ImageData.tm.nsec << "[nsec]" << std::endl;
		}

		// show image buffer
		if(m_ImageWindow == "on"){
			for (int i=0; i<2; ++i){
				_imgViewer[i].openWindow();
			}
			_imgViewer[0].display(&m_ImageData.data[0]);
			_imgViewer[1].waitKeyDisplay(&m_ImageDataR.data[0], 10);
		} else if(m_ImageWindow  == "off"){
			for (int i=0; i<2; ++i){
				_imgViewer[i].closeWindow();
			}
		}
	}else{// for single camera
		// capture image
		if(!_camera.CaptureImages((unsigned char*)&(m_ImageData.data[0]))){
			return RTC::RTC_ERROR;
		}

		// time stamp
		m_ImageData.tm = UTIL::getCurrentTimeStamp();

		// send image to outport
		if(!m_ImageDataOut.write()) {
			std::cout << "Image was not sent!" << std::endl;
			std::cout << "TimeStamp: " << m_ImageData.tm.sec << "[sec] " << m_ImageData.tm.nsec << "[nsec]" << std::endl;
		}

		// show image buffer
		if(m_ImageWindow == "on"){
			_imgViewer[0].openWindow();
			_imgViewer[0].waitKeyDisplay(&m_ImageData.data[0], 10);
		} else if(m_ImageWindow  == "off"){
			_imgViewer[0].closeWindow();
		}
	}
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t FireWebCamera::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t FireWebCamera::onError(RTC::UniqueId ec_id)
{
	std::cout << "error status!" << std::endl;
  return RTC::RTC_OK;
}



RTC::ReturnCode_t FireWebCamera::onReset(RTC::UniqueId ec_id)
{
	std::cout << "reset" << std::endl;
	// close camera
	if(_camera.getActivated()){
		_camera.Closes();
	}
	// close window
	for(int i=0; i<2; ++i){
		_imgViewer[i].closeWindow();
	}
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t FireWebCamera::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t FireWebCamera::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void FireWebCameraInit(RTC::Manager* manager)
  {
    Properties profile(firewebcamera_spec);
    manager->registerFactory(profile,
                             RTC::Create<FireWebCamera>,
                             RTC::Delete<FireWebCamera>);
  }
  
};


